# vim: tabstop=4 shiftwidth=4 softtabstop=4
#
# Copyright 2019 Eugene Frolov
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import contextlib
import logging

from restalchemy.storage.sql import engines


LOG = logging.getLogger(__name__)


class Context(object):

    def __init__(self, engine_name=engines.DEFAULT_NAME):
        super(Context, self).__init__()
        self._engine_name = engine_name

    def start_new_session(self):
        engine = self._engine
        storage = engine.get_session_storage()
        session = engine.get_session()
        storage.store_session(session)
        LOG.debug("New session %r has been started", session)
        return session

    @property
    def _engine(self):
        return engines.engine_factory.get_engine(name=self._engine_name)

    @contextlib.contextmanager
    def session_manager(self):
        session = self.start_new_session()
        try:
            yield session
            session.commit()
            LOG.debug("Session %r has been committed", session)
        except Exception:
            session.rollback()
            LOG.exception("Session %r has been rolled back by reason:",
                          session)
            raise
        finally:
            self.session_close()

    def _get_storage(self):
        engine = self._engine
        return engine.get_session_storage()

    def get_session(self):
        return self._get_storage().get_session()

    def session_close(self):
        session = self.get_session()
        try:
            session.close()
            LOG.debug("Session %r has been closed", session)
        except Exception:
            LOG.exception("Can't close session by reason:")
        finally:
            self._get_storage().remove_session()
            LOG.debug("Session %r has been removed from thread storage",
                      session)



# vim: tabstop=4 shiftwidth=4 softtabstop=4
#
# Copyright 2016 Eugene Frolov <eugene@frolov.net.ru>
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import threading
from wsgiref.simple_server import make_server
from wsgiref.simple_server import WSGIServer

from restalchemy.api import applications
from restalchemy.api.middlewares import (
    retry_on_error as retry_error_middleware)
from restalchemy.api.middlewares import errors as errors_middleware
from restalchemy.openapi import engines as openapi_engines
from restalchemy.openapi import structures as openapi_structures
from restalchemy.storage import exceptions as storage_exc
from restalchemy.storage.sql import engines
from restalchemy.tests.functional import consts
from restalchemy.tests.functional.restapi.ra_based.microservice import (
    middlewares)
from restalchemy.tests.functional.restapi.ra_based.microservice import routes


class RESTService(threading.Thread):

    def __init__(self, bind_host="127.0.0.1", bind_port=8080):
        super(RESTService, self).__init__(name="REST Service")

        openapi_engine = openapi_engines.OpenApiEngine(
            info=openapi_structures.OpenApiInfo(
                title='REST API Microservice',
                description='REST API Microservice for tests',
                version='1.2.3',
                contact=openapi_structures.OpenApiContact(
                    name='Functional Tests',
                    url='https://functional.tests/',
                    email='functional@tests.local',
                ),
                license=openapi_structures.OpenApiApacheLicense(),
                terms_of_service='https://functional.tests/terms/',
            ),
            paths=openapi_structures.OpenApiPaths(),
            components=openapi_structures.OpenApiComponents(),
            security=[openapi_structures.OpenApiSecurity()],
            tags=openapi_structures.OpenApiTags([openapi_structures.OpenApiTag(
                name='functional-test',
                description='Just functional tests',
                external_docs=openapi_structures.OpenApiExternalDocs(
                    url='https://https://functional.tests/docs/',
                    description='Functional tests external documentation'
                ),
            )]),
            external_docs=openapi_structures.OpenApiExternalDocs(
                url='https://https://functional.tests/docs/',
                description='Functional tests external documentation'
            ),
        )

        self._httpd = make_server(
            bind_host, bind_port,
            errors_middleware.ErrorsHandlerMiddleware(
                retry_error_middleware.RetryOnErrorsMiddleware(
                    middlewares.ContextMiddleware(
                        application=applications.OpenApiApplication(
                            route_class=routes.Root,
                            openapi_engine=openapi_engine
                        ),
                    ), exceptions=storage_exc.DeadLock,
                    # set max_retry == 2 to speed up retry tests execution
                    max_retry=2
                ),
            ),
            WSGIServer)

    def run(self):
        self._httpd.serve_forever()

    def stop(self):
        self._httpd.server_close()
        self._httpd.shutdown()
        self.join(timeout=10)


def main():
    rest_service = RESTService()
    try:
        rest_service.run()
    except KeyboardInterrupt:
        pass


if __name__ == "__main__":
    engines.engine_factory.configure_factory(consts.DATABASE_URI)
    main()



# vim: tabstop=4 shiftwidth=4 softtabstop=4
#
# Copyright 2016 Eugene Frolov <eugene@frolov.net.ru>
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import contextlib
import logging

from mysql.connector import pooling
import six
from six.moves.urllib import parse

from restalchemy.common import singletons
from restalchemy.storage.sql.dialect import adapters
from restalchemy.storage.sql.dialect import mysql
from restalchemy.storage.sql import sessions


DEFAULT_NAME = 'default'
DEFAULT_CONNECTION_TIMEOUT = 10
LOG = logging.getLogger(__name__)


@six.add_metaclass(abc.ABCMeta)
class AbstractEngine(object):

    @abc.abstractproperty
    def URL_SCHEMA(self):
        raise NotImplementedError()

    def __init__(self, db_url):
        super(AbstractEngine, self).__init__()
        self._db_url = parse.urlparse(db_url)

    @abc.abstractproperty
    def db_name(self):
        raise NotImplementedError()

    @abc.abstractproperty
    def db_username(self):
        raise NotImplementedError()

    @abc.abstractproperty
    def db_password(self):
        raise NotImplementedError()


class MySQLEngine(AbstractEngine):

    URL_SCHEMA = 'mysql'

    def __init__(self, db_url, config=None, query_cache=False):
        super(MySQLEngine, self).__init__(db_url)
        if self._db_url.scheme != self.URL_SCHEMA:
            raise ValueError("Database url should be starts with mysql://. "
                             "For example: mysql://username:password@"
                             "127.0.0.1:3306/database_name")
        config = config or {}
        self._db_name = self._db_url.path[1:]
        if 'connection_timeout' not in config:
            config['connection_timeout'] = DEFAULT_CONNECTION_TIMEOUT
        config.update({
            'user': self.db_username,
            'password': self.db_password,
            'database': self.db_name,
            'host': self.db_host,
            'port': self.db_port,
            'converter_class': adapters.MySQLConverter
        })

        try:
            self._pool = pooling.MySQLConnectionPool(**config)
        except AttributeError as e:
            pool_name = e.args[0].split("'")[1]
            new_name = str(hash(pool_name))
            LOG.warning("Changing '%s' pool name to '%s'",
                        pool_name, new_name)
            config["pool_name"] = new_name
            self._pool = pooling.MySQLConnectionPool(**config)

        self._dialect = mysql.MySQLDialect()
        self._session_storage = sessions.SessionThreadStorage()
        self._query_cache = query_cache

    def __del__(self):
        pool = getattr(self, '_pool', None)
        if pool is not None:
            self._pool._remove_connections()

    @property
    def query_cache(self):
        return self._query_cache

    @property
    def dialect(self):
        return self._dialect

    @property
    def db_name(self):
        return self._db_name

    @property
    def db_username(self):
        return self._db_url.username

    @property
    def db_password(self):
        return self._db_url.password

    @property
    def db_host(self):
        return self._db_url.hostname

    @property
    def db_port(self):
        return self._db_url.port or 3306

    def get_connection(self):
        return self._pool.get_connection()

    def get_session(self):
        return sessions.MySQLSession(engine=self)

    def _get_session_from_storage(self):
        try:
            return self.get_session_storage().get_session()
        except sessions.SessionNotFound:
            return None

    @contextlib.contextmanager
    def session_manager(self, session=None):
        session = session or self._get_session_from_storage()
        if session is None:
            session = self.get_session()
            try:
                yield session
                session.commit()
            except Exception:
                session.rollback()
                raise
            finally:
                session.close()
        else:
            yield session

    def get_session_storage(self):
        return self._session_storage


class EngineFactory(singletons.InheritSingleton):

    def __init__(self):
        super(EngineFactory, self).__init__()
        self._engines = {}
        self._engines_map = {
            MySQLEngine.URL_SCHEMA: MySQLEngine
        }

    def configure_factory(self, db_url, config=None, query_cache=False,
                          name=DEFAULT_NAME):
        """Configure_factory

        @property db_url: str. For example driver://user:passwd@host:port/db
        """
        schema = db_url.split(':')[0]
        try:
            self._engines[name] = self._engines_map[schema.lower()](
                db_url=db_url,
                config=config,
                query_cache=query_cache
            )
        except KeyError:
            raise ValueError("Can not find driver for schema %s" % schema)

    def get_engine(self, name=DEFAULT_NAME):
        engine = self._engines.get(name, None)
        if engine:
            return engine
        raise ValueError(("Can not return %s engine. Please configure"
                         " EngineFactory") % name)

    def destroy_engine(self, name=DEFAULT_NAME):
        try:
            del self._engines[name]
        except KeyError:
            pass

    def destroy_all_engines(self):
        self._engines = {}


class DBConnectionUrl(object):

    _CENSORED = ':<censored>@'

    def __init__(self, db_url):
        super(DBConnectionUrl, self).__init__()
        self._db_url = parse.urlparse(db_url)

    def __repr__(self):
        if self._db_url.password is None:
            orig_substr = "@"
        else:
            orig_substr = ":%s@" % self._db_url.password
        return self.url.replace(orig_substr, self._CENSORED)

    @property
    def url(self):
        return self._db_url.geturl()


engine_factory = EngineFactory()



# vim: tabstop=4 shiftwidth=4 softtabstop=4
#
# Copyright 2016 Eugene Frolov <eugene@frolov.net.ru>
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import json

import six

from restalchemy.common import exceptions as common_exc
from restalchemy.dm import filters as dm_filters
from restalchemy.dm import models
from restalchemy.storage import base
from restalchemy.storage import exceptions
from restalchemy.storage.sql.dialect import exceptions as exc
from restalchemy.storage.sql import engines
from restalchemy.storage.sql import tables


class ObjectCollection(base.AbstractObjectCollection,
                       base.AbstractObjectCollectionCountMixin):

    @property
    def _table(self):
        return self.model_cls.get_table()

    @property
    def _engine(self):
        return engines.engine_factory.get_engine()

    @base.error_catcher
    def get_all(self, filters=None, session=None, cache=False, limit=None,
                order_by=None, locked=False):
        with self._engine.session_manager(session=session) as s:
            if cache is True:
                return s.cache.get_all(
                    engine=self._engine,
                    table=self._table,
                    filters=filters,
                    fallback=self._get_all,
                    limit=limit,
                    order_by=order_by,
                    locked=locked,
                )

            return self._get_all(filters=filters, session=s, limit=limit,
                                 order_by=order_by, locked=locked)

    def _get_all(self, filters, session, limit, order_by=None, locked=False):
        result = self._table.select(
            engine=self._engine, filters=filters, limit=limit,
            order_by=order_by, session=session, locked=locked)
        return [self.model_cls.restore_from_storage(**params)
                for params in result.rows]

    @base.error_catcher
    def get_one(self, filters=None, session=None, cache=False, locked=False):
        result = self.get_all(
            filters=filters,
            session=session,
            cache=cache,
            limit=2,
            locked=locked,
        )
        result_len = len(result)
        if result_len == 1:
            return result[0]
        elif result_len == 0:
            raise exceptions.RecordNotFound(model=self.model_cls,
                                            filters=filters)
        else:
            raise exceptions.HasManyRecords(model=self.model_cls,
                                            filters=filters)

    def _query(self, where_conditions, where_values,
               session, limit, order_by, locked):
        result = self._table.custom_select(
            engine=self._engine,
            where_conditions=where_conditions,
            where_values=where_values,
            session=session,
            limit=limit,
            order_by=order_by,
            locked=locked,
        )
        return [self.model_cls.restore_from_storage(**params)
                for params in list(result.fetchall())]

    @base.error_catcher
    def query(self, where_conditions, where_values, session=None,
              cache=False, limit=None, order_by=None, locked=False):
        """

        :param where_conditions: "NOT (bala < %s)"
        :param where_values: (5, 10,)
        """
        with self._engine.session_manager(session=session) as s:
            if cache is True:
                return s.cache.query(
                    engine=self._engine,
                    table=self._table,
                    where_conditions=where_conditions,
                    where_values=where_values,
                    fallback=self._query,
                    limit=limit,
                    order_by=order_by,
                    locked=locked,
                )

            return self._query(where_conditions=where_conditions,
                               where_values=where_values,
                               session=s,
                               limit=limit,
                               order_by=order_by,
                               locked=locked)

    @base.error_catcher
    def count(self, session=None, filters=None):
        with self._engine.session_manager(session=session) as s:
            result = self._table.count(
                engine=self._engine, session=s, filters=filters)
            data = list(result.fetchall())
            return data[0]['COUNT']


class UndefinedAttribute(common_exc.RestAlchemyException):

    message = "Class attribute %(attr_name)s must be provided."


@six.add_metaclass(abc.ABCMeta)
class SQLStorableMixin(base.AbstractStorableMixin):

    _saved = False

    _ObjectCollection = ObjectCollection

    __tablename__ = None

    @classmethod
    def get_table(cls):
        try:
            table = cls.__operational_storage__.get(
                tables.OPERATIONAL_STORAGE_SIMPLE_TABLE_KEY,
            )
        except common_exc.NotFoundOperationalStorageError:
            if cls.__tablename__ is None:
                raise UndefinedAttribute(attr_name='__tablename__')
            table = tables.SQLTable(engine=cls._get_engine(),
                                    table_name=cls.__tablename__,
                                    model=cls)
            cls.__operational_storage__.store(
                tables.OPERATIONAL_STORAGE_SIMPLE_TABLE_KEY,
                table,
            )
        return table

    @classmethod
    def _get_engine(cls):
        return engines.engine_factory.get_engine()

    @classmethod
    def restore_from_storage(cls, **kwargs):
        model_format = {}
        for name, value in kwargs.items():
            model_format[name] = (cls.properties.properties[name]
                                  .get_property_type()
                                  .from_simple_type(value))
        obj = cls.restore(**model_format)
        obj._saved = True
        return obj

    @base.error_catcher
    @base.dead_lock_catcher
    def insert(self, session=None):
        # TODO(efrolov): Add filters parameters.
        with self._get_engine().session_manager(session=session) as s:
            try:
                self.get_table().insert(engine=self._get_engine(),
                                        data=self._get_prepared_data(),
                                        session=s)
                # TODO(efrolov): Check result
            except exc.Conflict as e:
                raise exceptions.ConflictRecords(model=self, msg=str(e))
            self._saved = True

    def save(self, session=None):
        # TODO(efrolov): Add filters parameters.
        self.update(session) if self._saved else self.insert(session)

    @base.error_catcher
    @base.dead_lock_catcher
    def update(self, session=None, force=False):
        # TODO(efrolov): Add filters parameters.
        if self.is_dirty() or force:
            self.validate()
            with self._get_engine().session_manager(session=session) as s:
                try:
                    result = self.get_table().update(
                        engine=self._get_engine(),
                        ids=self._get_prepared_data(self.get_id_properties()),
                        data=self._get_prepared_data(
                            self.get_data_properties()),
                        session=s)
                except exc.Conflict as e:
                    raise exceptions.ConflictRecords(model=self, msg=str(e))
                if result.get_count() == 0:
                    _filters = {
                        name: dm_filters.EQ(prop.value)
                        for name, prop in self.get_id_properties().items()}
                    type(self).objects.get_one(filters=_filters, session=s)
                if result.get_count() > 1:
                    raise exceptions.MultipleUpdatesDetected(model=self,
                                                             filters={})

    @base.error_catcher
    @base.dead_lock_catcher
    def delete(self, session=None):
        # TODO(efrolov): Add filters parameters.
        with self._get_engine().session_manager(session=session) as s:
            result = self.get_table().delete(
                engine=self._get_engine(),
                ids=self._get_prepared_data(self.get_id_properties()),
                session=s)
            # TODO(efrolov): Check result
            return result

    @classmethod
    def to_simple_type(cls, value):
        if value is None:
            return None
        id_type = cls.get_id_property().popitem()[-1].get_property_type()
        if isinstance(value, models.Model):
            return id_type.to_simple_type(value.get_id())
        # Allow to filter by id without full model
        return id_type.to_simple_type(value)

    @classmethod
    @base.dead_lock_catcher
    def from_simple_type(cls, value):
        if value is None:
            return None
        if isinstance(value, base.PrefetchResult):
            for name in cls.id_properties.keys():
                if value[name]:
                    break
            else:
                return None
            return cls.restore_from_storage(**value)
        for name in cls.id_properties:
            value = (cls.properties.properties[name].get_property_type()
                        .from_simple_type(value))
            engine = engines.engine_factory.get_engine()
            return cls.objects.get_one(filters={name: dm_filters.EQ(value)},
                                       cache=engine.query_cache)


@six.add_metaclass(abc.ABCMeta)
class SQLStorableWithJSONFieldsMixin(SQLStorableMixin):

    __jsonfields__ = None

    @classmethod
    def restore_from_storage(cls, **kwargs):
        if cls.__jsonfields__ is None:
            raise UndefinedAttribute(attr_name='__jsonfields__')
        kwargs = kwargs.copy()
        for field in cls.__jsonfields__:
            kwargs[field] = json.loads(kwargs[field])
        return super(SQLStorableWithJSONFieldsMixin, cls
                     ).restore_from_storage(**kwargs)

    def _get_prepared_data(self, properties=None):
        if self.__jsonfields__ is None:
            raise UndefinedAttribute(attr_name='__jsonfields__')
        result = super(SQLStorableWithJSONFieldsMixin, self
                       )._get_prepared_data(properties)
        if properties is None:
            json_properties = self.__jsonfields__
        else:
            json_properties = set(self.__jsonfields__).intersection(
                set(properties.keys()))
        for field in json_properties:
            result[field] = json.dumps(result[field])
        return result



#!/usr/bin/env python
#
# Copyright (c) 2020 Eugene Frolov
# Copyright (c) 2018 Eugene Frolov
# Copyright (c) 2014 Eugene Frolov <eugene@frolov.net.ru>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from logging import config as logging_config

from oslo_config import cfg
import yaml


DEFAULT_CONFIG = {
    'version': 1,
    'formatters': {
        'aardvark': {
            'datefmt': '%Y-%m-%d,%H:%M:%S',
            'format': "%(asctime)15s.%(msecs)03d %(processName)s"
                      " pid:%(process)d tid:%(thread)d %(levelname)s"
                      " %(name)s:%(lineno)d %(message)s"
        }
    },
    'handlers': {
        'console': {
            'class': 'logging.StreamHandler',
            'formatter': 'aardvark',
            'stream': 'ext://sys.stderr'
        },
    },
    'loggers': {
        'restalchemy': {},
    },
    'root': {
        'level': 'DEBUG',
        'handlers': ['console']
    }
}


logging_opts = [
    cfg.StrOpt('config', default='logging.yaml',
               help="Logging subsystem configuration YAML file")
]

cfg.CONF.register_cli_opts(logging_opts, 'logging')


class ConfigNotFound(Exception):
    pass


def configure():
    config = cfg.CONF.logging.config
    config_file = cfg.CONF.find_file(config)

    if config_file is None:
        config_data = DEFAULT_CONFIG
    else:
        with open(config_file) as f:
            config_data = yaml.safe_load(f)

    logging_config.dictConfig(config_data)

    if config_data == DEFAULT_CONFIG:
        logging.getLogger(__name__).warning(
            'Logging configuration %s not found - using defaults',
            config)
